__author__ = 'Paul Gundarapu'
__author_email__ = 'leogps@outlook.com'
__version__ = '0.1.14'
__license__ = 'MIT'
__description__ = 'Simple, zero-configuration command-line static HTTP server.'
__long_description__ = """
Sipper is a simple, zero-configuration command-line static HTTP server. It is built on bottle:
https://github.com/bottlepy/bottle | https://bottlepy.org/docs/dev/ It aims to provide the same value provided by
http-server tool written in nodejs: https://github.com/http-party/http-server |
https://www.npmjs.com/package/http-server

Homepage and documentation: https://github.com/leogps/bottle-sipper

Copyright (c) 2023, Paul Gundarapu. License: MIT (see LICENSE for details)
"""



import unittest
import os
from sipper import Sipper


class TestApplication(unittest.TestCase):
    """Checking if command-line args work"""

    def test_directory(self):
        sipper = Sipper('/')
        self.assertEqual(sipper.directory, '/')

    def test_show_directory_listings(self):
        sipper = Sipper('/')
        self.assertEqual(sipper.show_directory_listings, True)

    def test_shutdown(self):
        sipper = Sipper('/')
        sipper.start_sipping('0.0.0.0', 8089)
        sipper.shutdown(wait_before_shutdown=2)

    def test_ssl(self):
        test_dir = os.path.dirname(os.path.abspath(__file__))
        cert = os.path.join(test_dir, 'test-cert', 'server.crt')
        key = os.path.join(test_dir, 'test-cert', 'server.key')

        sipper = Sipper('/', ssl_enabled=True, ssl_cert=cert, ssl_key=key)
        sipper.start_sipping('0.0.0.0', 8099)
        sipper.shutdown(wait_before_shutdown=2)

    def test_ssl_validation(self):
        with self.assertRaises(Exception):
            Sipper('/', ssl_enabled=True)

        test_dir = os.path.dirname(os.path.abspath(__file__))

        with self.assertRaises(Exception):
            cert = os.path.join(test_dir, 'test-cert', 'server.crt')
            Sipper('/', ssl_enabled=True, ssl_cert=cert)

        with self.assertRaises(Exception):
            key = os.path.join(test_dir, 'test-cert', 'server.key')
            Sipper('/', ssl_enabled=True, ssl_key=key)


if __name__ == "__main__":
    unittest.main()



import gzip
from io import BytesIO


def apply_gzip(request, response, uncompressed_response):
    accept_encoding = request.headers.get('Accept-Encoding', '')
    if 'gzip' in accept_encoding:
        # Compress the response data
        gzip_buffer = BytesIO()
        with gzip.GzipFile(mode='wb', compresslevel=6, fileobj=gzip_buffer) as f:
            f.write(uncompressed_response.encode('utf-8'))

        compressed_data = gzip_buffer.getvalue()

        # Add the 'Content-Encoding' header to indicate compression
        response.headers['Content-Encoding'] = 'gzip'
        response.headers['Content-Length'] = len(compressed_data)
        return compressed_data
    return uncompressed_response



import json
import os
import sys

import pkg_resources

YES = ['true', '1', 't', 'y', 'yes', 'yeah', 'yup']

APP_NAME = 'bottle-sipper'

APP_LINK = 'https://github.com/leogps/bottle-sipper'


def load_json(json_file_path):
    with open(json_file_path, 'r') as json_file:
        return json.load(json_file)


#
# Loading json files into variables.
#
_current_script_path = os.path.abspath(__file__)

# Get the directory of the current script
_current_script_dir = os.path.dirname(_current_script_path)

_mime_type_extensions_json_file = os.path.join(_current_script_dir, '../static', 'mime_type_extensions.json')
_mime_extension_map = load_json(_mime_type_extensions_json_file)

_icons_json_file = os.path.join(_current_script_dir, '../static', 'icons.json')
_icons = load_json(_icons_json_file)

#
# Loading version
#
_python_version = sys.version


class Template:
    def __init__(self, name, path):
        self.name = name
        self.path = path


#
# Initializing templates
#
_default_template = Template(name='default', path=pkg_resources.resource_filename(__name__, 'templates'
                                                                                            '/default/'))
_media_template = Template(name='media', path=pkg_resources.resource_filename(__name__, 'templates/media/'))
_templates = [
    _default_template,
    _media_template
]


def get_icons():
    return _icons


def get_mime_extensions():
    return _mime_extension_map


def get_python_version():
    return _python_version


def get_templates():
    return _templates


def get_default_template():
    return _default_template


def find_existing_template(template_name):
    for t in get_templates():
        if t.name == template_name:
            return t
    return None



from bottle import ServerAdapter
import ssl


class SipperWSGIRefServer(ServerAdapter):
    """ Custom WSGIRefServer implementation to accommodate shutdown, ssl etc. """
    server = None

    def __init__(self, host='127.0.0.1', port=8080,
                 ssl_enabled=False,
                 ssl_cert=None,
                 ssl_key=None,
                 silent=False,
                 **options):
        super().__init__(host, port, **options)
        self.ssl_enabled = ssl_enabled
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key
        self.silent = silent

    def run(self, handler):
        from wsgiref.simple_server import make_server, WSGIRequestHandler
        if self.quiet:
            class QuietHandler(WSGIRequestHandler):
                if self.silent:
                    def log_request(*arguments, **kw): pass
                else:
                    pass

            self.options['handler_class'] = QuietHandler
        self.server = make_server(self.host, self.port, handler, **self.options)

        if self.ssl_enabled:
            # Configure SSL context with certificates
            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(certfile=self.ssl_cert, keyfile=self.ssl_key)

            # Wrap the server with SSL/TLS context
            self.server.socket = ssl_context.wrap_socket(
                self.server.socket,
                server_side=True
            )
        self.server.serve_forever()

    def shutdown(self):
        self.server.server_close()
        self.server.shutdown()



from abc import abstractmethod


class Authentication:
    def __init__(self, enabled=False, mechanism=None):
        self.enabled = enabled
        self.mechanism = mechanism
        self.failed_status_code = 401
        self.failed_message = 'Access Denied.'

    @abstractmethod
    def decorate_with_auth_required(self, res):
        pass

    @abstractmethod
    def authenticate(self, header):
        pass


class BasicAuthentication(Authentication):

    def __init__(self, realm='', enabled=False, username=None, password=None):
        Authentication.__init__(self, enabled=enabled, mechanism='Basic')
        self.username = username
        self.password = password
        self.realm = realm
        self.__header_regex = r'Basic (?P<credentials>.*)'
        self.__username_password_regex = r'(?P<username>.*):(?P<password>.*)'

    def decorate_with_auth_required(self, res):
        res.set_header('WWW-Authenticate', 'Basic realm="%s"' % self.realm)
        res.set_header('Content-Type', 'text/html')
        res.status = self.failed_status_code
        res.body = self.failed_message

    def authenticate(self, header):
        import re
        basic_header_match = re.match(self.__header_regex, header)
        if basic_header_match is None:
            return False
        credentials = basic_header_match.group('credentials')
        from base64 import b64decode
        decoded = b64decode(credentials).decode('utf-8')
        username_password_match = re.match(self.__username_password_regex, decoded)
        if username_password_match is None:
            return False
        username = username_password_match.group('username')
        password = username_password_match.group('password')
        return self.username == username and self.password == password



import hashlib
import os
from sipper_core.constants import get_icons


def perms_to_string(stat, is_dir):
    directory = 'd' if is_dir else '-'
    mode = oct(stat.st_mode)[-3:]

    permissions = [
        '---', '--x', '-w-', '-wx', 'r--', 'r-x', 'rw-', 'rwx'
    ]

    permission_string = ''.join([permissions[int(n, 8)] for n in mode])

    return directory + permission_string


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def handle_shortcut_symbols(directory):
    if os.name == 'nt' and directory.startswith('~'):
        return directory.replace('~', os.path.expanduser('~'), 1)
    return directory


def build_icons():
    icons = []
    icons_json = get_icons()
    for ext in icons_json.keys():
        icon = Icon(name=ext, base64_data=icons_json[ext])
        icons.append(icon)
    return icons


def file_exists_and_is_file(file_path):
    return os.path.exists(file_path) and os.path.isfile(file_path)


class Icon:
    def __init__(self, name, base64_data):
        self.name = name
        self.base64_data = base64_data


class FileDetails:
    def __init__(self,
                 is_dir,
                 file_icon_style_class,
                 file_icon_base64,
                 last_modified_date,
                 file_permissions,
                 file_size,
                 file_link,
                 file_name):
        self.is_dir = is_dir
        self.file_icon_style_class = file_icon_style_class
        self.file_icon_base64 = file_icon_base64
        self.last_modified_date = last_modified_date
        self.file_permissions = file_permissions
        self.file_size = file_size
        self.file_link = file_link
        self.file_name = file_name
        b = ':'.join([
            self.file_name,
            str(self.is_dir),
            self.file_permissions,
            str(self.file_size),
            str(is_dir)
        ]).encode('utf-8')
        hash_digest = hashlib.sha256(b).hexdigest()
        self.hash = hash_digest

    def to_json(self):
        return {
            "hash": self.hash,
            "isDir": self.is_dir,
            "fileIconStyleClass": self.file_icon_style_class,
            "fileIconBase64": self.file_icon_base64,
            "lastModifiedDate": self.last_modified_date,
            "filePermissions": self.file_permissions,
            "fileSize": self.file_size,
            "fileLink": self.file_link,
            "fileName": self.file_name
        }



from distutils.core import setup
from setuptools import find_packages

from sipper_core.metadata import __version__
from sipper_core.metadata import __license__
from sipper_core.metadata import __description__
from sipper_core.metadata import __long_description__
from sipper_core.metadata import __author__
from sipper_core.metadata import __author_email__


setup(
    name='bottle-sipper',
    py_modules=[
        'sipper'
    ],
    packages=find_packages(where="."),
    package_dir={
      "": "."
    },
    package_data={
        "sipper_core.templates.default": [
          "*.tpl",
          "*.html",
          "*.js"
        ],
        "sipper_core.templates.media": [
            "*.tpl",
            "*.html",
            "*.js"
        ],
        "static": ["*.json"],
        "test": [
            "test-cert/*.crt",
            "test-cert/*.key"
        ]
    },
    include_package_data=True,
    version=__version__,
    license=__license__,
    description=__description__,
    long_description=__long_description__,
    long_description_content_type='text/markdown',
    author=__author__,
    author_email=__author_email__,
    url='https://github.com/leogps/bottle-sipper',
    keywords=['bottle',
              'http',
              'static',
              'file',
              'server',
              'http-server'],
    install_requires=[
        'bottle>=0.12.23',
        'argparse>=1.4.0',
        'ifaddr>=0.2.0'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Sipper is a simple, zero-configuration command-line static HTTP server. It is built on bottle:
https://github.com/bottlepy/bottle | https://bottlepy.org/docs/dev/ It aims to provide the same value provided by
http-server tool written in nodejs: https://github.com/http-party/http-server |
https://www.npmjs.com/package/http-server

Homepage and documentation: https://github.com/leogps/bottle-sipper

Copyright (c) 2023, Paul Gundarapu. License: MIT (see LICENSE for details)
"""
import json
import os
import pathlib
from argparse import ArgumentParser
from datetime import datetime
from threading import Thread
from time import sleep

import ifaddr
from bottle import get, run, static_file, request, response, HTTPResponse, SimpleTemplate
from bottle import __version__ as bottle_version

from sipper_core.server import SipperWSGIRefServer
from sipper_core.auth import Authentication, BasicAuthentication
from sipper_core.constants import get_icons, get_mime_extensions, YES, get_templates, get_default_template, \
    find_existing_template, \
    get_python_version, \
    APP_NAME, \
    APP_LINK
from sipper_core.common import perms_to_string, sizeof_fmt, handle_shortcut_symbols, build_icons, \
    file_exists_and_is_file, FileDetails
from sipper_core.compress import apply_gzip
from sipper_core.metadata import __version__


class Sipper(Thread):

    def __init__(self,
                 directory,
                 show_directory_listings=True,
                 auth=Authentication(),
                 use_available_template=get_default_template().name,
                 template_base_dir=None,
                 ssl_enabled=False,
                 ssl_cert=None,
                 ssl_key=None,
                 searchable=False,
                 gzip=False,
                 silent=False):
        Thread.__init__(self)
        directory = handle_shortcut_symbols(directory)
        self.directory = directory
        self.show_directory_listings = show_directory_listings
        self.authentication = auth
        self.daemon = False
        self.threads = []
        self.servers = []
        # Handling ssl config
        if ssl_enabled:
            if ssl_cert is None or ssl_key is None:
                raise Exception('ssl cert/key missing. Need ssl cert and ssl key to enable ssl.')
            elif not file_exists_and_is_file(ssl_cert):
                raise Exception('ssl_cert file does not exist or is not a file.')
            elif not file_exists_and_is_file(ssl_key):
                raise Exception('ssl_key file does not exist or is not a file.')

        self.ssl_enabled = ssl_enabled
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key
        # Handling templates.
        if template_base_dir is not None:
            self.template_base_dir = template_base_dir
        else:
            if use_available_template is None:
                use_available_template = get_default_template().name
            existing_template = find_existing_template(use_available_template)
            if existing_template is None:
                raise Exception('template not found: ' + use_available_template + '\nAvailable templates: ' +
                                ', '.join(
                                    i.name
                                    for i in
                                    get_templates()
                                ))
            else:
                self.template_base_dir = existing_template.path
        self.searchable = searchable
        self.gzip = gzip
        self.silent = silent

    def get_server_address(self):
        host = request.get_header('Host')
        if host is None:
            hostname = ''
        else:
            hostname = host.split(':')[0]
        return ':'.join([hostname, str(self.server_port)])

    def handle_auth(self, req, res):
        header = req.get_header('Authorization')
        if not isinstance(header, str) or not self.authentication.authenticate(header):
            self.authentication.decorate_with_auth_required(res)
            raise res.copy(cls=HTTPResponse)

    def serve(self, url_path=''):
        if self.authentication.enabled:
            self.handle_auth(request, response)
        # print('Requested: %s' % url_path)
        url_path_normalized = url_path.replace('/', '', 1)
        filename = os.path.join(self.directory, url_path_normalized)
        is_dir = os.path.isdir(filename)
        if is_dir:
            return self.handle_dir(filename)

        if self.gzip:
            gz_file = url_path_normalized + '.gz'
            gzip_file_full_path = os.path.join(self.directory, gz_file)
            if file_exists_and_is_file(gzip_file_full_path):
                return self.handle_gzip_file(gz_file=gz_file)

        extension = pathlib.Path(filename).suffix
        ext = extension.lower().replace('.', '')
        mime_type = None
        mime_type_extensions = get_mime_extensions()
        if ext in mime_type_extensions:
            mime_type = mime_type_extensions[ext]
        return static_file(filename=url_path_normalized, root=self.directory, mimetype=mime_type)

    def handle_gzip_file(self, gz_file):
        full_gz_file_path = os.path.join(self.directory, gz_file)
        with open(full_gz_file_path, 'rb') as binary_file:
            binary_data = binary_file.read()
        response.headers['Content-Encoding'] = 'gzip'
        response.headers['Content-Length'] = len(binary_data)
        return binary_data

    def handle_dir(self, path):
        if os.path.isfile(os.path.join(path, 'index.html')):
            return static_file(filename='index.html', root=path)

        if not self.show_directory_listings:
            return HTTPResponse(status=403, body=None)

        dir_list = sorted(os.listdir(path), key=lambda v: (not os.path.isdir(os.path.join(path, v)), v.upper()))

        file_details_list = []
        icons = build_icons()
        icons_json = get_icons()
        for f in dir_list:
            file_path = os.path.join(path, f)

            icon_style_class = '_blank'
            if os.path.isfile(file_path):
                extension = pathlib.Path(file_path).suffix
                ext = extension.lower().replace('.', '')
                if ext in icons_json:
                    icon_style_class = ext
            file_details_file_icon_base64 = icons_json[icon_style_class]

            last_modified = os.path.getmtime(file_path)
            modified_date = datetime.fromtimestamp(last_modified).strftime('%d-%b-%Y %H:%M')

            stat_info = os.stat(file_path)

            file_permissions = ''.join(['(', perms_to_string(stat_info, os.path.isdir(file_path)), ')'])

            if os.path.isfile(file_path):
                file_details_is_dir = False
                size = stat_info.st_size
                size_formatted = sizeof_fmt(size)
            else:
                file_details_is_dir = True
                size_formatted = ''

            link_path = '/'.join([path.replace(self.directory, '', 1), f]).replace('//', '/')
            link_path = ('/' + link_path) if not link_path.startswith('/') else link_path

            file_details = FileDetails(is_dir=file_details_is_dir,
                                       file_icon_style_class=icon_style_class,
                                       file_icon_base64=file_details_file_icon_base64,
                                       last_modified_date=modified_date,
                                       file_permissions=file_permissions,
                                       file_size=size_formatted,
                                       file_link=link_path,
                                       file_name=f)
            file_details_list.append(file_details)

        html = SimpleTemplate(name=self.build_template_file_path('index.html'))
        index_of = path.replace(self.directory, '', 1)

        html_model = {
            'dir': index_of,
            'template_base_dir': self.template_base_dir,
            'file_details_list': file_details_list,
            'icons': icons,
            'python_version': get_python_version(),
            'app_name': APP_NAME,
            'app_version': __version__,
            'app_link': APP_LINK,
            'server_address': self.get_server_address(),
            'searchable': self.searchable
        }
        if self.searchable:
            html_model['file_details_json'] = json.dumps([detail.to_json() for detail in file_details_list])

        html = html.render(**html_model)
        response.content_type = 'text/html'
        if self.gzip:
            html = apply_gzip(request=request, response=response, uncompressed_response=html)
        return html

    def build_template_file_path(self, template_file):
        return os.path.join(self.template_base_dir, template_file)

    def _run(self, address, port,
             ssl_enabled=False,
             ssl_cert=None,
             ssl_key=None):
        # print('Serving at http://{}:{}'.format(ip, port))
        server = SipperWSGIRefServer(host=address,
                                     port=port,
                                     ssl_enabled=ssl_enabled,
                                     ssl_cert=ssl_cert,
                                     ssl_key=ssl_key,
                                     silent=self.silent)
        self.servers.append(server)
        self.server_port = port
        print("Bottle v%s server starting up (using %s)...\n" % (bottle_version, repr(server)))
        protocol = 'https' if ssl_enabled else 'http'
        print("Starting server on %s://%s:%d/\n" % (protocol, server.host, server.port))
        run(quiet=True, server=server)

    def start_sipping(self, address, port):
        thread = Thread(target=self._run, kwargs={
            'address': address,
            'port': port,
            'ssl_enabled': self.ssl_enabled,
            'ssl_cert': self.ssl_cert,
            'ssl_key': self.ssl_key
        })
        thread.start()
        self.threads.append(thread)

    def shutdown(self, wait_before_shutdown=0):
        print('Shutting down...')
        sleep(wait_before_shutdown)
        for server in self.servers:
            server.shutdown()

    def config_formatted(self):
        config = 'Show Directory Listings: %s' % self.show_directory_listings
        if self.authentication.enabled:
            config += '\n' + 'Authentication Enabled: True'
        if self.ssl_enabled:
            config += '\n' + 'SSL Enabled: True'
        for t in get_templates():
            if t.path == self.template_base_dir:
                config += '\n' + 'Template: ' + t.name
                break
        return config


def main():
    pass


if __name__ == "__main__":
    main()

    parser = ArgumentParser()
    parser.add_argument('-d', '--show-dir', default=True, required=False, help='Show directory listings')
    parser.add_argument('-a', '--address', required=False, help='Address for the server, defaults to 0.0.0.0')
    parser.add_argument('-p', '--port', default=8080, required=False, help='Port for the server')
    # auth options
    auth_arg_group = parser.add_argument_group('auth-options')
    auth_arg_group.add_argument('-u', '--username', default=None, required=False,
                                help='Username for basic authentication')
    auth_arg_group.add_argument('-P', '--password', default=None, required=False,
                                help='Password for basic authentication')

    parser.add_argument('-b', '--template-base-dir', default=None, required=False,
                        help='Template base directory. '
                             'Takes precedence over --use-available-template option')
    parser.add_argument('-t', '--use-available-template', default=None, required=False,
                        help='Use out-of-the-box templates. '
                             'Available templates: ' + ', '.join(i.name for i in get_templates()))
    # ssl options
    ssl_arg_group = parser.add_argument_group('ssl-options')
    ssl_arg_group.add_argument('-S', '--ssl-enabled', '--tls-enabled', default=False, required=False,
                               action='store_true',
                               help='Enable secure request serving with TLS/SSL (HTTPS).')
    ssl_arg_group.add_argument('-C', '--cert', default=None, required=False, help='Path to ssl cert file')
    ssl_arg_group.add_argument('-K', '--key', default=None, required=False, help='Path to ssl key file')

    parser.add_argument('-q', '--searchable', action='store_true', default=False, required=False,
                        help='Add search box to be able to search on files (Performs fuzzy search similar to fzf tool).')
    parser.add_argument('-g', '--gzip', action='store_true', default=False, required=False,
                        help='When enabled, it will server some-file.js.gz file in place of some-file.js when a '
                             'gzipped version of the file exists and the request accepts gzip encoding.'
                             ' Also applies gzip to the directory listing response.')
    parser.add_argument('-s', '--silent', action='store_true', default=False, required=False,
                        help='Suppress log messages from output')

    parser.add_argument('directory')

    args = parser.parse_args()
    if not isinstance(args.show_dir, bool):
        args.show_dir = True if (args.show_dir in YES) else False

    authentication = Authentication()
    if args.username is not None or args.password is not None:
        if args.username is None:
            args.username = ''
        if args.password is None:
            args.password = ''
        authentication = BasicAuthentication(enabled=True, username=args.username, password=args.password)

    if args.ssl_enabled and (not args.cert or not args.key):
        parser.error('--cert and --key are required to enable ssl.')

    sipper = Sipper(args.directory,
                    show_directory_listings=args.show_dir,
                    auth=authentication,
                    use_available_template=args.use_available_template,
                    template_base_dir=args.template_base_dir,
                    ssl_enabled=args.ssl_enabled,
                    ssl_cert=args.cert,
                    ssl_key=args.key,
                    searchable=args.searchable,
                    gzip=args.gzip,
                    silent=args.silent)
    get('<url_path:path>')(sipper.serve)

    print('Starting up bottle-sipper, serving %s' % sipper.directory)
    print('')
    print('bottle-sipper version: %s' % __version__)
    print('')
    print('bottle-sipper settings:')
    print(sipper.config_formatted())
    print('')

    if args.address is None:
        adapters = ifaddr.get_adapters()
        for adapter in adapters:
            # print("IPs of network adapter " + adapter.nice_name)
            for ip in adapter.ips:
                # print("   %s/%s" % (ip.ip, ip.network_prefix))
                theIp = ip.ip
                if (theIp is None or (not isinstance(theIp, str)) or (ip.is_IPv6 and theIp.find('%') != -1) or (
                        ip.nice_name.startswith('utun')
                        or (ip.nice_name.startswith('bridge'))
                        or ip.nice_name.startswith('virbr')
                )):
                    continue
                sipper.start_sipping(theIp, args.port)
    else:
        sipper.start_sipping(args.address, args.port)
        pass
